from .plotly_plot.plotly_plot import plotly_plot
'''
A statistical Python toolbox to explore brain signals
'''

class Channel():
	'Channels of Brain Signals\' epochs to help performing exploratory data analysis (EDA)'
	def __init__(self, epoch, channel_id):
		self.epoch = epoch
		self.dataset = Channel.get_channel(epoch, channel_id)
		self.name = epoch.name
		self.sampling_rate = epoch.sampling_rate
		self.num_channels = epoch.num_channels
		self.epoch_length = epoch.epoch_length
		self.channel_id = channel_id

	def __repr__(self): # when the user calls the object information of the channel is printed
		return self.__str__()

	def __str__(self): # when the user calls the object information of the channel is printed
		return 'Channel {}'.format(self.channel_id)

	def plot(self, bands=None):
		band_set = ['delta','theta','alpha','beta','gamma']
		if bands == 'all':
			specified_bands = band_set
		else:
			specified_bands = [element for element in band_set if element in bands] if bands is not None else []
		
		lines = [self.dataset]
		names = ['Raw Data']

		for band in specified_bands:
			arr = self.__getattr__(band)
			if arr is not None:
				lines.append(arr)
				names.append(band)
		
		title = '{} - {}'.format(self.epoch.__str__(), self.__str__())
		plotly_plot.line_chart(names, title, *lines)


	def __getattr__(self, key):
		if key in ['delta', 'theta', 'alpha', 'beta', 'gamma']:
			try:
				return self.epoch.__getattr__(key)[:,self.channel_id-1]
			except TypeError:
				return None
	
	@staticmethod
	def get_channel(epoch, channel_id):
		'Given an Epoch object, it returns the data frame for the specified channel number'
		return epoch.dataset[:,channel_id-1]


# -*- coding: utf-8 -*-

"""Top-level package for Python Brain Signal Analysis."""

__author__ = """Abdulrahman A. Althobaiti"""
__email__ = 'Abdulrahman.Althobaiti@kaust.edu.sa'
__version__ = '0.1.0'

#from ._version import get_versions
#__version__ = get_versions()['version']
#del get_versions
from .loader import loader as loader


import sys
from time import sleep

import numpy as np

def fancy_print(string):
	if string[-1] != '\n':
		string = string + '\n'
	for char in string:
		sys.stdout.write('{}'.format(char))
		sys.stdout.flush()
		sleep(0.025)

def printList(lst):
	for i, val in enumerate(lst):
		print('{}) {}'.format(i+1, val))

def array_reshape_by_column(array, num_columns_index):
	if len(array.shape) > 2:
		new_array = []
		for col in range(array.shape[num_columns_index]):
			new_array.append(array[:,col].flatten())
		return np.array(new_array)
	else:
		return np.transpose(array)

def normalize_channels(array2d):
	return np.divide(np.subtract(array2d, array2d.mean(axis=0)), array2d.std(axis=0))

def verify_list_str_within_bounds(low, high, list_or_str):
	new_list = []

	if isinstance(list_or_str, list):
		pass
	elif isinstance(list_or_str, str):
		list_or_str = list_or_str.split()
	else:
		raise ValueError('Expected str or list of number(s), got {}:{}'.format(
			type(list_or_str), list_or_str))

	for element in list_or_str:
		if int(element) >= low and int(element) <= high:
			new_list.append(int(element))
		else:
			raise ValueError('Expected values between {} and {}, got {}'.format(
				low, high, element))

	return new_list

def ccf(x, y, maxlags=26):
	Nx = len(x)
	c = np.correlate(x, y, mode=2)
	c/= np.sqrt(np.dot(x,x) * np.dot(y,y))
	c = c[Nx-1-maxlags:Nx+maxlags]
	return c

'''
This is another version of ccf with fft, providing more percesion.
However, it is not fast. that is why it is commented out.

import numpy as np
from scipy.fftpack import fft, fftshift, ifft, fftn

def ccf_shot(x, y, Nx, maxlags=26):
	return np.real(fftshift(ifft(np.multiply(x, np.conj(y)))))[Nx-1-maxlags:Nx+maxlags]

def ccf_driver(frame):
	for ch_1 in range(len(frame[0])):
		for ch_2 in range(ch_1, len(frame[0])):
			ccf(frame[:,ch_1], frame[:,ch_2])

def ccf_fft_driver(arg):
	Nx = len(arg[:,0])
	corr_length = Nx * 2 - 1
	#frame_1 = fftn(frame, shape=(corr_length,), axes=[0])
	for ch_1 in range(len(frame[0])):
		x = fft(arg[:,ch_1], corr_length)
		for ch_2 in range(ch_1, len(frame[0])):
			ccf_shot(x, fft(arg[:,ch_2], corr_length), Nx)
'''


from .channel import Channel
from .plotly_plot.plotly_plot import plotly_plot
from .utilities import verify_list_str_within_bounds

'''
A statistical Python toolbox to explore brain signals

TODO:
xcorr plot
	if single band use interactive mode
	if dual band use matplotlib
'''

class Epoch():
	'Segmentation of Brain Signals to help performing exploratory data analysis (EDA)'
	def __init__(self, brain_signals, epoch_id):
		self.brain_signals = brain_signals
		self.dataset = Epoch.get_epoch(brain_signals, epoch_id)
		self.name = brain_signals.name
		self.sampling_rate = brain_signals.sampling_rate
		self.num_channels = brain_signals.num_channels
		self.epoch_length = brain_signals.epoch_length
		self.epoch_id = epoch_id
		self.channels = [Channel(self, channel_id) for channel_id in range(1, self.num_channels + 1)]
		self.xcorr_dict = {}
		self.lags_dict = {}

	def __getitem__(self, index): 
		'gives access to channels where indexing starts from 1 to length of channels.'
		if index < 1 or index > len(self.channels):
			raise ValueError('Index should be between 1 and {}.'.format(len(self.channels)))
		return self.channels[index-1]

	def __len__(self):
		'returns the number of channels in the epoch'
		return len(self.channels)

	def __repr__(self):
		return self.__str__()

	def __str__(self):
		return 'Epoch {}'.format(self.epoch_id)
	
	def __getattr__(self, key):
		if key in ['delta', 'theta', 'alpha', 'beta', 'gamma']:
			row_lower_bound = (self.epoch_id - 1) * self.sampling_rate
			row_upper_bound = self.epoch_id * self.sampling_rate
			try:
				return self.brain_signals.__getattr__(key)[row_lower_bound:row_upper_bound,:]
			except TypeError:
				return None

	def plot(self, band='raw', channels=None, exclude=False):
		dataset = None

		if band not in ['raw', 'delta','theta','alpha','beta','gamma']:
			raise ValueError('{} is not an acceptable parameter.\n'\
							'Allowed parameters are: \'raw\', \'delta\','\
							'\'theta\',\'alpha\',\'beta\', or \'gamma\''.format(band))
		elif band == 'raw':
			dataset = self.dataset
		else:
			dataset = self.__getattr__(band)
		
		if dataset is None:
			print('The {} dataset is empty, cannot proceed!'.format(band))
			return

		channels = verify_list_str_within_bounds(1, self.num_channels, channels)\
					if channels is not None else None
		lines=[]
		names=[]
		if channels is not None:
			for ch in range(self.num_channels):
				if exclude and ch+1 not in channels:
					lines.append(dataset[:,ch])
					names.append('Ch {}'.format(ch+1))
				elif not exclude and ch+1 in channels:
					lines.append(dataset[:,ch])
					names.append('Ch {}'.format(ch+1))
		else:
			for ch in range(self.num_channels):
				lines.append(dataset[:,ch])
				names.append('Ch {}'.format(ch+1))

		plotly_plot.line_chart(names, '{}'.format(self.__str__()), *lines)
		
	@staticmethod
	def get_epoch(brain_signals, epoch_id):
		'''
		Given a brain-signal object, it returns the data frame for the specified
		epoch number
		'''
		return brain_signals.dataset[((epoch_id - 1) \
			   * brain_signals.sampling_rate): \
			   (epoch_id * brain_signals.sampling_rate),:]


import numpy as np
from scipy.signal.sigtools import _linear_filter

def filtfilt(b, a, x, zi):
    axis = -1
    max_a_b = max(len(a), len(b))
    # edge, ext = _validate_pad(padtype, padlen, x, axis, ntaps=max(len(a), len(b)))
    # unfloded below
    edge = max_a_b * 3
    # left_end = axis_slice(x, start=0, stop=1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(0, 1, None)
    left_end = x[a_slice]

    # left_ext = axis_slice(x, start=edge, stop=0, step=-1, axis=axis)
    # unfolded below
    #-> ignore a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(edge, 0, -1)
    left_ext = x[a_slice]

    # right_end = axis_slice(x, start=-1, axis=axis)
    # unfolded below
    #-> ignore a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(-1, None, None)
    right_end = x[a_slice]

    # right_ext = axis_slice(x, start=-2, stop=-(edge + 2), step=-1, axis=axis)
    # unfolded below
    #-> ignore a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(-2, -(edge +2), -1)
    right_ext = x[a_slice]

    ext = np.concatenate((2 * left_end - left_ext, x, 2 * right_end - right_ext), axis=axis)
  
    # NEW UNFOLDED SECTION
    zi_shape = [1] * x.ndim
    zi_shape[axis] = zi.size
    zi = np.reshape(zi, zi_shape)
    # x0 = axis_slice(ext, stop=1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * ext.ndim
    a_slice[axis] = slice(None, 1, None)
    x0 = ext[a_slice]

    # NEW UNFOLDED SECTION
    # (y, zf) = lfilter(b, a, ext, axis=axis, zi=zi * x0)
    # unfolded with
    (y, zf) = _linear_filter(b, a, ext, axis, zi * x0)

    # NEW UNFOLDED SECTION
    # y0 = axis_slice(y, start=-1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * y.ndim
    a_slice[axis] = slice(-1, None, None)
    y0 = y[a_slice]

    # NEW UNFOLDED SECTION
    # y1 = axis_reverse(y, axis=axis)
    # unfolded with
    a_slice = [slice(None)] * y.ndim
    a_slice[axis] = slice(None, None, -1)
    y1 = y[a_slice]

    # NEW UNFOLDED SECTION
    # (y, zf) = lfilter(b, a, y1, axis=axis, zi=zi * y0)
    # unfolded below
    (y, zf) = _linear_filter(b, a, y1, axis, zi * y0)

    # NEW UNFOLDED SECTION
    # y = axis_reverse(y, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * y.ndim
    a_slice[axis] = slice(None, None, -1)
    y = y[a_slice]

    if edge > 0:
        # NEW UNFOLDED SECTION
        # y = axis_slice(y, start=edge, stop=-edge, axis=axis)
        # unfolded below
        a_slice = [slice(None)] * y.ndim
        a_slice[axis] = slice(edge, -edge, None)
        y = y[a_slice]

    return y

'''
import numpy as np
from scipy.signal.signaltools import _validate_pad, lfilter_zi, lfilter
from scipy.signal._arraytools import axis_reverse, axis_slice
from scipy.signal.sigtools import _linear_filter


# axis_slice:
# a_slice = [slice(None)] * {name_of_caller_arg}.ndim
# a_slice[axis] = slice(None, None, None)
# {name_of_result} = {name_of_caller_arg}[a_slice]

def filtfilt_unfolded(b, a, x, axis=-1, padtype='odd', padlen=None, method='pad', irlen=None):
    # edge, ext = _validate_pad(padtype, padlen, x, axis, ntaps=max(len(a), len(b)))
    # unfloded below
    max_a_b = max(len(a), len(b))
    edge = max_a_b * 3
    # left_end = axis_slice(x, start=0, stop=1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(0, 1, None)
    left_end = x[a_slice]

    # left_ext = axis_slice(x, start=edge, stop=0, step=-1, axis=axis)
    # unfolded below
    #-> ignore a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(edge, 0, -1)
    left_ext = x[a_slice]

    # right_end = axis_slice(x, start=-1, axis=axis)
    # unfolded below
    #-> ignore a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(-1, None, None)
    right_end = x[a_slice]

    # right_ext = axis_slice(x, start=-2, stop=-(edge + 2), step=-1, axis=axis)
    # unfolded below
    #-> ignore a_slice = [slice(None)] * x.ndim
    a_slice[axis] = slice(-2, -(edge +2), -1)
    right_ext = x[a_slice]

    ext = np.concatenate((2 * left_end - left_ext, x, 2 * right_end - right_ext), axis=axis)
  
    # NEW UNFOLDED SECTION
    # This section can be computed only once for each band, and passed along as a parameter
    # zi = lfilter_zi(b, a)
    # unfolded below

    # linalg.companion(a)
    # unfolded below
    first_row = -a[1:] / (1.0 * a[0])
    n = a.size
    c = np.zeros((n - 1, n - 1), dtype=first_row.dtype)
    c[0] = first_row
    c[list(range(1, n - 1)), list(range(0, n - 2))] = 1

    IminusA = np.eye(max_a_b - 1) - c.T
    B = b[1:] - a[1:] * b[0]
    # Solve zi = A*zi + B
    zi = np.linalg.solve(IminusA, B)

    # NEW UNFOLDED SECTION
    zi_shape = [1] * x.ndim
    zi_shape[axis] = zi.size
    zi = np.reshape(zi, zi_shape)
    # x0 = axis_slice(ext, stop=1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * ext.ndim
    a_slice[axis] = slice(None, 1, None)
    x0 = ext[a_slice]

    # NEW UNFOLDED SECTION
    # (y, zf) = lfilter(b, a, ext, axis=axis, zi=zi * x0)
    # unfolded with
    (y, zf) = _linear_filter(b, a, ext, axis, zi * x0)

    # NEW UNFOLDED SECTION
    # y0 = axis_slice(y, start=-1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * y.ndim
    a_slice[axis] = slice(-1, None, None)
    y0 = y[a_slice]

    # NEW UNFOLDED SECTION
    # y1 = axis_reverse(y, axis=axis)
    # unfolded with
    a_slice = [slice(None)] * y.ndim
    a_slice[axis] = slice(None, None, -1)
    y1 = y[a_slice]

    # NEW UNFOLDED SECTION
    # (y, zf) = lfilter(b, a, y1, axis=axis, zi=zi * y0)
    # unfolded below
    (y, zf) = _linear_filter(b, a, y1, axis, zi * y0)

    # NEW UNFOLDED SECTION
    # y = axis_reverse(y, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * y.ndim
    a_slice[axis] = slice(None, None, -1)
    y = y[a_slice]

    if edge > 0:
        # NEW UNFOLDED SECTION
        # y = axis_slice(y, start=edge, stop=-edge, axis=axis)
        # unfolded below
        a_slice = [slice(None)] * y.ndim
        a_slice[axis] = slice(edge, -edge, None)
        y = y[a_slice]

    return y

def filtfilt_original(b, a, x, axis=-1, padtype='odd', padlen=None, method='pad', irlen=None):

    # method == "pad"
    edge, ext = _validate_pad(padtype, padlen, x, axis, ntaps=max(len(a), len(b)))

    # Get the steady state of the filter's step response.
    zi = lfilter_zi(b, a)

    # Reshape zi and create x0 so that zi*x0 broadcasts
    # to the correct value for the 'zi' keyword argument
    # to lfilter.
    zi_shape = [1] * x.ndim
    zi_shape[axis] = zi.size
    zi = np.reshape(zi, zi_shape)
    x0 = axis_slice(ext, stop=1, axis=axis)

    # Forward filter.
    (y, zf) = lfilter(b, a, ext, axis=axis, zi=zi * x0)

    # Backward filter.
    # Create y0 so zi*y0 broadcasts appropriately.
    y0 = axis_slice(y, start=-1, axis=axis)
    (y, zf) = lfilter(b, a, axis_reverse(y, axis=axis), axis=axis, zi=zi * y0)

    # Reverse y.
    y = axis_reverse(y, axis=axis)

    if edge > 0:
        # Slice the actual signal from the extended signal.
        y = axis_slice(y, start=edge, stop=-edge, axis=axis)

    return y

def filtfilt(b, a, x, axis=-1, padtype='odd', padlen=None, method='pad', irlen=None):
#    filtfilt(b, a, x, axis=-1, padtype='odd', padlen=None, irlen=None)

    # method == "pad"
    # edge, ext = _validate_pad('odd', None, x, -1, ntaps=max(len(a), len(b)))
    # unfolded below

    max_a_b = max(len(a), len(b))

    edge = max_a_b * 3

    a_slice = [slice(None)] * x.ndim
    a_slice[-1] = slice(0, 1, None)
    left_end = x[a_slice]
    a_slice[-1] = slice(edge, 0, -1)
    left_ext = x[a_slice]
    a_slice[-1] = slice(-1, None, None)
    right_end = x[a_slice]
    a_slice[-1] = slice(-2, -(edge + 2), -1)
    right_ext = x[a_slice]

    ext = np.concatenate((2 * left_end - left_ext,
                          x,
                          2 * right_end - right_ext),
                         axis=-1)

    # Get the steady state of the filter's step response.
    # zi = lfilter_zi(b, a)
    # unfolded below

    # linalg.companion(a)
    # unfolded below
    first_row = -a[1:] / (1.0 * a[0])
    n = a.size
    c = np.zeros((n - 1, n - 1), dtype=first_row.dtype)
    c[0] = first_row
    c[list(range(1, n - 1)), list(range(0, n - 2))] = 1

    IminusA = np.eye(max_a_b - 1) - c.T
    B = b[1:] - a[1:] * b[0]
    # Solve zi = A*zi + B
    zi = np.linalg.solve(IminusA, B)

    # Reshape zi and create x0 so that zi*x0 broadcasts
    # to the correct value for the 'zi' keyword argument
    # to lfilter.
    zi_shape = [1] * x.ndim
    zi_shape[axis] = zi.size
    zi = np.reshape(zi, zi_shape)
    # x0 = axis_slice(ext, stop=1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * ext.ndim
    a_slice[-1] = slice(None, 1, None)
    x0 = ext[a_slice]

    # Forward filter.
    (y, zf) = lfilter(b, a, ext, axis=axis, zi=zi * x0)

    # Backward filter.
    # Create y0 so zi*y0 broadcasts appropriately.
    # y0 = axis_slice(y, start=-1, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * y.ndim
    a_slice[-1] = slice(-1, None, None)
    y0 = y[a_slice]

    # y1 = axis_reverse(y, axis=axis)
    # unfolded below
    a_slice = [slice(None)] * y.ndim
    a_slice[-1] = slice(None, None, -1)
    y1 = y[a_slice]

    (y, zf) = lfilter(b, a, y1, axis=axis, zi=zi * y0)

    # Reverse y.
    # y = axis_reverse(y, axis=axis)
    a_slice = [slice(None)] * y.ndim
    a_slice[-1] = slice(None, None, -1)
    y = y[a_slice]

    if edge > 0:
        # Slice the actual signal from the extended signal.
        # y = axis_slice(y, start=edge, stop=-edge, axis=axis)
        # unfolded below
        a_slice = [slice(None)] * y.ndim
        a_slice[-1] = slice(edge, edge, None)
        y = y[a_slice]

    return y
'''


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

#import versioneer
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'numpy>=1.11.1',
    'scipy>=0.18.1',
    'plotly>=2.0.12',
    'h5py>=2.6.0'
]

setup_requirements = [
    # TODO(A-Althobaiti): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='prainsa',
#    version=versioneer.get_version(),
    description="A Python framework to analyze and visualize brain signals.",
    long_description=readme + '\n\n' + history,
    author="Abdulrahman A. Althobaiti",
    author_email='Abdulrahman.Althobaiti@kaust.edu.sa',
    url='https://github.com/A-Althobaiti/prainsa',
    packages=find_packages(include=['prainsa']),
    include_package_data=True,
    install_requires=requirements,
    license="GNU General Public License v3",
    zip_safe=False,
    keywords='prainsa',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `prainsa` package."""


import unittest, os, prainsa

from prainsa import loader


class test_prainsa(unittest.TestCase):
    """Tests for `prainsa` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.mat_v73 = os.path.dirname(os.path.abspath(prainsa.__file__)) + '/data/matlab_array_v73.mat'
        self.mat_v6 = os.path.dirname(os.path.abspath(prainsa.__file__)) + '/data/matlab_array_v6.mat'
        self.numpy_array = os.path.dirname(os.path.abspath(prainsa.__file__)) + '/data/numpy_array.npy'
        self.zipped_numpy_array = os.path.dirname(os.path.abspath(prainsa.__file__)) + '/data/zipped_numpy_array.npz'
        self.prainsa_object = os.path.dirname(os.path.abspath(prainsa.__file__)) + '/data/prainsa_object'

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def test_loading_v73_mat_file(self):
        loader(self.mat_v73, None, 'example_data', 15, 100, 1)

    def test_loading_v6_mat_file(self):
        loader(self.mat_v6, None, 'example_data', 15, 100, 1)

    def test_loading_numpy_array_file(self):
        loader(self.numpy_array, None, 'example_data', 15, 100, 1)

    def test_loading_zipped_numpy_array_file(self):
        loader(self.zipped_numpy_array, None, 'example_data', 15, 100, 1)

    def test_loading_prainsa_object(self):
        loader(self.prainsa_object)

    def test_confirm_mat_file_numpy_file(self):
        v73 = loader(self.mat_v73, None, 'example_data', 15, 100, 1)
        np_array = loader(self.numpy_array, None, 'example_data', 15, 100, 1)
        assert(sum(sum(v73.dataset == np_array.dataset)) == (15 * 5000))

    def test_confirm_mat_file_zipped_numpy_file(self):
        v73 = loader(self.mat_v73, None, 'example_data', 15, 100, 1)
        zipped_np_array = loader(self.numpy_array, None, 'example_data', 15, 100, 1)
        assert(sum(sum(v73.dataset == zipped_np_array.dataset)) == (15 * 5000))

    def test_confirm_mat_file_prainsa_object(self):
        v73 = loader(self.mat_v73, None, 'example_data', 15, 100, 1)
        prainsa_object = loader(self.prainsa_object)
        assert(sum(sum(v73.dataset == prainsa_object.dataset)) == (15 * 5000))

    def test_confirm_mat_file_older_version(self):
        v73 = loader(self.mat_v73, None, 'example_data', 15, 100, 1)
        v6 = loader(self.mat_v6, None, 'example_data', 15, 100, 1)
        assert(sum(sum(v73.dataset == v6.dataset)) == (15 * 5000))


import os
import errno
from os.path import isdir as dir_exist
from os.path import isfile as file_exist
from os import getcwd as pwd
from os import mkdir
from time import time

import numpy as np
from scipy.io import loadmat
from scipy.signal import butter, resample
from scipy.stats import zscore

from .utilities import fancy_print, ccf, normalize_channels
from .speed_up import filtfilt
from .epoch import Epoch
from .plotly_plot.plotly_plot import plotly_plot

'''
A statistical Python toolbox to explore brain signals
'''

class Prainsa(object):
	'A toolbox to explore brain signals connectivity'

	def __init__(self, dataset, name, sampling_rate=1000, num_channels=32, epoch_length=1):
		'''
		dataset: holds the actual brain signals, either as list or a numpy array, where columns are channels and rows are observations
		name: specifies the name of the dataset
		sampling_rate: is pre-defined as 1000; however, it must be overridden -specified- if the sampling rate is different
		num_channels: is pre-defined as 32; however, it must be overriden -specified- if the number of channels in the dataset is different
		epoch_length: is an arbitrary segmentatoin of the data, and the default value is 1 second
		'''
		self.dataset = dataset
		self.name = name
		self.sampling_rate = sampling_rate
		self.num_channels = num_channels
		self.epoch_length = epoch_length
		self.validate_input()
		self.epochs = [Epoch(self, epoch_id) for epoch_id in range(1, int(self.dataset.shape[0] / (self.sampling_rate * self.epoch_length)) + 1)]
		self.xcorr_dict = {}
		self.lags_dict = {}

	def __getitem__(self, index): 
		'gives access to epochs where indexing starts from 1 to length of epochs.'
		if index < 1 or index > len(self.epochs):
			raise ValueError('Index should be between 1 and {}.'.format(len(self.epochs)))
		return self.epochs[index-1]

	def __len__(self):
		'returns the number of segmentations, aka. epochs'
		return len(self.epochs)

	def __repr__(self):
		return 'BrainSignals(dataset={}, name={}, sampling_rate={}, num_channels={}, epoch_length={})'.format(self.dataset, self.name, self.sampling_rate, self.num_channels, self.epoch_length)

	def __str__(self):
		return 'Brain Signals object\n--------------------\n{}\nDataset dimensions: {}\nSampling rate: {} Hertz\nNumber of channels: {} channels\nEpoch length: {} seconds'.format(self.name, self.dataset.shape, self.sampling_rate, self.num_channels, self.epoch_length)

	def __getattr__(self, key):
		if key in ['delta', 'theta', 'alpha', 'beta', 'gamma']:
			if key in self.__dict__:
				return self.__dict__[key]
			else:
				fancy_print('The {} band is not extracted.\nRun the extract method with \'{}\' as an argument.'.format(key, key))
				return None
		raise KeyError('{} is not a valid attribute'.format(key))

	def validate_input(self):
		'checks if the dimensions of the data comply with the number of channels and sampling rate'
		nrow, ncol = self.dataset.shape[0], self.dataset.shape[1]
		if ncol != self.num_channels:
			raise ValueError('Number of columns must match number of channels')
		if nrow % self.sampling_rate != 0:
			raise ValueError('Number of rows must be a multiple of sampling rate')

	def extract(self, band):
		'''
		Extracts a frequency band of the channels for each epoch.

		Parameters
		----------
		band : takes one or more of the following frequency bands
				+---------+--------------------+
				| 'delta' | 0.5 Hz to 4 Hz     |
				+---------+--------------------+
				| 'theta' | 4 Hz to 8 Hz       |
				+---------+--------------------+
				| 'alpha' | 8 Hz to 12 Hz      |
				+---------+--------------------+
				| 'beta'  | 12 Hz to 30 Hz     |
				+---------+--------------------+
				| 'gamma' | 30 Hz to 100 Hz    |
				+---------+--------------------+
				|  'all'  | extracts all bands |
				+---------+--------------------+

				the ranges of the bands are pre-defined and
				cannot be modeified. Unless the frequency range
				exceeds the sampling rate of the dataset, then
				the upper limit will be replaced by the sampling
				rate - 1.

		Returns
		----------
		None. The resulting computations will be stored in the object
		and accessible through calling '{object_name}.{band_name}'.
		'''
		bands = ['delta','theta','alpha','beta','gamma']
		frequency_range = {'delta':[0.5,4],'theta':[4,8],'alpha':[8,12],'beta':[12,30],'gamma':[30,100]}
		
		band_set = []

		start = time()

		if not band == 'all':
			band_set = [element for element in bands if element in band] if band is not None else []
		else:
			band_set = bands
			
		if band_set == []:
			raise ValueError('Acceptable band names are: \'delta\', \'theta\', \'alpha\', \'beta\', and \'gamma\'.')

		for band in band_set:
			low, high = frequency_range[band]
			if low >= self.sampling_rate:
				raise ValueError(errno.EINVAL, os.strerror(errno.EINVAL), '{} low frequency {} cannot be greater than the sampling rate {}.'.format(band, low, self.sampling_rate))
			if high >= self.sampling_rate:
				fancy_print('[Warning] {} high frequency {} is greater than the sampling rate {}, and it will be replaced with {}.'.format(band, low, self.sampling_rate, self.sampling_rate/2))
				high = self.sampling_rate/2
			self.__dict__[band] = np.zeros(self.dataset.shape, dtype=np.float32)
			# DO NOT MODIFY #
			# THIS IS MEANT FOR OPTIMIZATION, MUST NOT BE MODIFIED #
			b, a = butter(N=3, Wn=(2 * low / self.sampling_rate, 2 * high / self.sampling_rate), btype='bandpass', analog=False, output='ba')
			first_row = -a[1:] / (1.0 * a[0])
			n = a.size
			c = np.zeros((n - 1, n - 1), dtype=first_row.dtype)
			c[0] = first_row
			c[list(range(1, n - 1)), list(range(0, n - 2))] = 1
			IminusA = np.eye(max(len(a), len(b)) - 1) - c.T
			B = b[1:] - a[1:] * b[0]
			zi = np.linalg.solve(IminusA, B)
			#########################################################
			for epoch in self.epochs:
				for channel in epoch.channels:
					row_lower_bound = (epoch.epoch_id - 1) * self.sampling_rate
					row_upper_bound = epoch.epoch_id * self.sampling_rate
					column = channel.channel_id-1
					self.__dict__[band][row_lower_bound:row_upper_bound, column] = filtfilt(b, a, channel.dataset, zi)

		end = time()
		total_time = end - start
		minutes = int(total_time / 60)
		seconds = int(total_time % 60)
		ms = int(total_time * 1000 % 1000)
		fancy_print('Finished extracting {}.\nExtraction time: {} minute(s), {} second(s), and {} millisecond(s)'.format(band_set, minutes, seconds, ms))

	def xcorr(self, first_band, second_band=None, maxlags=26):
		'''
		Performs a cross-correlation between pairs of channels 
		over all epochs for a given oscillatory band(s).

		Parameters
		----------
		first_band: specify the name or names for the bands to be
					cross-correlated. It could be a string or a list.
					Allowed parameters are:
					'delta','theta','alpha','beta','gamma', or 'all'

		second_band: (optional) specify the name or names for the bands
					 to perform a dual-frequency cross-correlation. It
					 could be a string or a list.
					 Allowed parameters are:
					 'delta','theta','alpha','beta','gamma', or 'all'

		maxlags: specify the number of maximum lags to be considered.
				 It must be an integer.

		Returns
		-------
		None. The resulting computations will be stored in the object
		and accessible through calling:
		'{object_name}.display_xcorr({first_band_name}, {second_band_name})'
		where the second_band_name is optional. It is intended for dual-
		frequency cross-correlation.

		Notes
		-----
		if .xcorr is called with a first_band only, it will perform regular
		cross-correlation on that band. Likewise, if the second_band is the
		same as the first_band, then a regular cross-correlation on the band
		will be performed.

		Warning
		-------
		The percesion of this computation has been decreased for both memory
		and computation optimization. Therefore, a value of zero, 0, might
		be represented as 0.00000016e-16.
		'''
		start = time()
		bands = ['delta','theta','alpha','beta','gamma']
		lags = np.arange(-maxlags, maxlags+1)
		first_band_set, second_band_set = [], []

		if not first_band == 'all':
			first_band_set = [element for element in bands if element in first_band] if first_band is not None else []
		else:
			first_band_set = bands
		
		if not second_band == 'all':
			second_band_set = [element for element in bands if element in second_band] if second_band is not None else []
		else:
			second_band_set = bands
		
		if first_band_set == []:
			raise ValueError('Acceptable band names are: delta, theta, alpha, beta, and gamma.')

		dual_bands = []

		if second_band_set == []:
			dual_bands = ['{}-{}'.format(band, band) for band in first_band_set]
		else:
			dual_bands = ['{}-{}'.format(band_1, band_2) for band_1 in first_band_set for band_2 in second_band_set]

		for dual_band in dual_bands:
			band_1, band_2 = dual_band.split('-')
			xcorr_values = np.zeros((len(self), self.num_channels, self.num_channels), dtype=np.float32)
			lags_values = np.zeros((len(self), self.num_channels, self.num_channels), dtype=np.float32)
			for epoch_index in range(len(self)):
				for channel_1_index in range(self.num_channels):
					for channel_2_index in range(channel_1_index, self.num_channels):
						row_lower_bound = epoch_index * self.sampling_rate
						row_upper_bound = (epoch_index + 1) * self.sampling_rate
						first_channel = None
						second_channel = None
						try:
							first_channel = self.__dict__[band_1][row_lower_bound:row_upper_bound, channel_1_index]
							second_channel = self.__dict__[band_2][row_lower_bound:row_upper_bound, channel_2_index]
						except KeyError:
							fancy_print('The band ({} and {}) must be extracted first. Run the \'extract\' method.'.format(band_1, band_2))
							return None
						ac = ccf(first_channel, second_channel, maxlags)
						ac = ac ** 2
						max_index = np.argmax(ac)
						xcorr_values[epoch_index, channel_1_index, channel_2_index] = ac[max_index]
						xcorr_values[epoch_index, channel_2_index, channel_1_index] = ac[max_index]
						lags_values[epoch_index, channel_1_index, channel_2_index] = lags[max_index]
						lags_values[epoch_index, channel_2_index, channel_1_index] = -lags[max_index]
			self.xcorr_dict[dual_band] = xcorr_values
			self.lags_dict[dual_band] = lags_values
		
		end = time()
		total_time = end - start
		minutes = int(total_time / 60)
		seconds = int(total_time % 60)
		ms = int(total_time * 1000 % 1000)
		
		fancy_print('Finished cross-correlation of {}.\nCross-correlation time: {} minute(s), {} second(s), and {} millisecond(s)'.format(dual_bands, minutes, seconds, ms))

	def boxplot_channels(self):
		normalized_channels = normalize_channels(self.dataset)
		absolute_normalized_channels = np.abs(normalized_channels)
		#ignored 
		#absolute_normalized_channels = normalized_channels

		
		plotly_plot.boxplot(y=absolute_normalized_channels.mean(axis=0),
							name='Channels',
							title='{}'.format(self.name))

	def boxplot_epochs(self):
		resampled_dataset = resample(self.dataset, len(self.epochs), axis=0)
		z_scores = zscore(resampled_dataset, axis=0, ddof=0)

		plotly_plot.boxplot(y=z_scores.mean(axis=1),
							name='Epoch',
							title='{}'.format(self.name))

	def save(self, path='./Brain Signals/', overwrite=False):
		'''
		saves the current object to disk at Brain Signals directory, which is located at the
		current working directory if not specified; in other words, './Brain Signals/'.
		Each object is saved into its own directory with the name specified upon initialization
		of Brain Signals object.
		'''

		def save_metadata():
			start = time()
			list_of_metadata = ['name','sampling_rate','num_channels','epoch_length']
			with open('{}{}/metadata.dat'.format(path, self.name),'w') as metadata:
				for element in list_of_metadata:
					metadata.write('{}={}\n'.format(element, self.__dict__[element]))
			end = time()
			return end - start

		def save_dataset():
			start = time()
			np.savez_compressed('{}{}/dataset'.format(path, self.name), dataset=self.dataset)
			end = time()
			return end - start

		def save_bands():
			start = time()
			bands_dict = {}
			bands = ['delta','theta','alpha','beta','gamma']
			for band in bands:
				if band in self.__dict__:
					bands_dict[band] = self.__dict__[band]
			if len(bands_dict) > 0:
				np.savez_compressed('{}{}/bands'.format(path, self.name), **bands_dict)
			end = time()
			return end - start

		def save_xcorr():
			start = time()
			if len(self.xcorr_dict) > 0:
				np.savez_compressed('{}{}/xcorr'.format(path, self.name), **self.xcorr_dict)
			end = time()
			return end - start

		def save_lags():
			start = time()
			if len(self.xcorr_dict) > 0:
				np.savez_compressed('{}{}/lags'.format(path, self.name), **self.lags_dict)
			end = time()
			return end - start

		saving_time = 0

		if not dir_exist('{}{}'.format(path, self.name)):
			mkdir('{}{}'.format(path, self.name))
		else:
			if not overwrite:
				fancy_print('A BrainSignals object with the name: {} already exists,'.format(self.name))
				fancy_print('at {}{}'.format(path, self.name))
				ans = None
				while ans != 'Y' and ans != 'N':
					fancy_print('Would you like to overwrite it?')
					ans = input('[Y|N]: ')
				if ans == 'N':
					return

		saving_time += save_metadata()
		saving_time += save_dataset()
		saving_time += save_bands()
		saving_time += save_xcorr()
		saving_time += save_lags()
		minutes = int(saving_time / 60)
		seconds = int(saving_time % 60)
		ms = int(saving_time * 1000 % 1000)
		fancy_print('Finished saving BrainSignals object.\nSaving time: {} minute(s), {} second(s), and {} millisecond(s)'.format(minutes, seconds, ms))
